//****************************************************************************
//**
//**    VFS_DEMO.CPP
//**    No Description
//**
//**	Project:	VFS
//**	Component:	DEMO
//**    Author:		Michael Walter
//**
//**	History:
//**		06.07.2001		Created (Michael Walter)
//****************************************************************************

//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#include <VFS/VFS.h>
#include <direct.h>
#include <algorithm>
#include <cassert>
#include <cstdio>
#include <conio.h>

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
static class CSampleFilter : public VFS_Filter
{
	VFS_UINT m_uAmount;

public:
	// Constructor / Destructor.
	CSampleFilter()
	{ m_uAmount = 1; }
	CSampleFilter( VFS_UINT uAmount )
	{ m_uAmount = uAmount; }
	~CSampleFilter()
	{}

	// Encoding / Decoding Procedures.
	VFS_BOOL Encode( VFS_FilterReadProc Reader, VFS_FilterWriteProc Writer, const VFS_EntityInfo& DecodedInfo ) const
	{ VFS_BYTE bByte; while( Reader( &bByte, 1 ) ) Writer( &(bByte+=m_uAmount), 1 ); return VFS_TRUE; }

	VFS_BOOL Decode( VFS_FilterReadProc Reader, VFS_FilterWriteProc Writer, const VFS_EntityInfo& EncodedInfo ) const
	{ VFS_BYTE bByte; while( Reader( &bByte, 1 ) ) Writer( &(bByte-=m_uAmount), 1 ); return VFS_TRUE; }

	// Filter Configuration Data Management.
	VFS_BOOL LoadConfigData( VFS_FilterReadProc Reader )
	{ return Reader( ( VFS_BYTE* )&m_uAmount, sizeof( VFS_UINT ) ); }

	VFS_BOOL SaveConfigData( VFS_FilterWriteProc Writer ) const
	{ return Writer( ( const VFS_BYTE* )&m_uAmount, sizeof( VFS_UINT ) ); }

	VFS_DWORD GetConfigDataSize() const
	{ return sizeof( VFS_UINT ); }

	// Information.
	VFS_PCSTR GetName() const
	{ return L"Sample Filter"; }

	VFS_PCSTR GetDescription() const
	{ return L"This is a sample Filter that just adds a Number to the Data for encoding."; }
} g_SampleFilter( 2 );

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static VFS_UINT g_uNumDirs = 0, g_uNumFiles = 0, g_uNumBytes = 0;

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static VFS_String GetDataDirPath();
static void ReportError();
static VFS_BOOL Shower( const VFS_EntityInfo& Info, void* pParam );

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
// Returns a sample ".\\Data" Directory.
static VFS_String GetDataDirPath()
{
	static wchar_t szBuffer[ 1024 ];

	_wgetcwd( szBuffer, 1024 );

	// Get the Module File Name.
	VFS_String strPath = szBuffer;

	// Remove the "Bin"/"Debug"/"Release" Directory.
	assert( strPath.rfind( VFS_TEXT( '\\' ) ) != std::string::npos );
	strPath = strPath.substr( 0, strPath.rfind( VFS_TEXT( '\\' ) ) );

	// Add our "Data" Directory.
	strPath = strPath +  VFS_TEXT( "\\Data" );

	// Let's use the VFS_PATH_SEPARATOR as the path separator.
	std::replace( strPath.begin(), strPath.end(), VFS_TEXT( '\\' ), VFS_PATH_SEPARATOR );

	return strPath;
}

static void ReportError()
{
	VFS_ErrorCode eError = VFS_GetLastError();
	wprintf( L"VFS Error: %s (%u)", VFS_GetErrorString( eError ), eError );
	getch();
	exit( 1 );
}

static VFS_BOOL Shower( const VFS_EntityInfo& Info, void* pParam )
{
	wprintf( L"> %s\t%s\t%u\t%s\t%s <\n", Info.strName.c_str(), Info.strPath.c_str(), Info.lSize, ( Info.eType == VFS_FILE ? L"file" : ( Info.eType == VFS_DIR ? L"dir" : L"archive" ) ), Info.bArchived ? L"(archived)" : L"" );
	g_uNumDirs += Info.eType == VFS_DIR ? 1 : 0;
	g_uNumFiles += Info.eType == VFS_FILE ? 1 : 0;
	g_uNumBytes += Info.lSize;
	return VFS_TRUE;
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================
// The Program Entry Point.
int main( int argc, char** argv )
{
	// Initialize the VFS.
	VFS_BOOL bSuccess = VFS_Init();
	assert( bSuccess );

	// Say "Hello".
	wprintf( L"VFS %u.%u\n", VFS_HIBYTE( VFS_GetVersion() ), VFS_LOBYTE( VFS_GetVersion() ) );

	// Root path management.
	VFS_AddRootPath( GetDataDirPath() );

	// Register a Sample Filter.
	VFS_RegisterFilter( &g_SampleFilter );

    VFS_FilterNameList Filters;
	/*Filters.push_back( g_SampleFilter.GetName() );
	Filters.push_back( g_SampleFilter.GetName() );*/
	if( !VFS_Archive_CreateFromDirectory( VFS_TEXT( "testy" ), VFS_TEXT( "test" ), Filters, VFS_TRUE ) )
		ReportError();

    VFS_EntityInfo Info;
	if( !VFS_File_GetInfo( VFS_TEXT( "testy/neuer ordner/kopie (2) von neuer ordner/kopie (6) von test.txt" ), Info ) )
		ReportError();
	Shower( Info, NULL );

	VFS_Handle hFile = VFS_File_Open( Info.strPath, VFS_READ );
	if( hFile == VFS_INVALID_HANDLE_VALUE )
		ReportError();
	char* pBuffer = new char[ Info.lSize + 1 ];
	if( !VFS_File_Read( hFile, ( VFS_BYTE* )pBuffer, Info.lSize ) )
		ReportError();
	pBuffer[ Info.lSize ] = '\0';
	printf( "---\nFile Contents (%u):\n---\n%s\n---\n", Info.lSize, pBuffer );
	delete[] pBuffer;
	if( !VFS_File_Close( hFile ) )
		ReportError();

    g_uNumDirs = 0, g_uNumFiles = 0, g_uNumBytes = 0;
	if( !VFS_Dir_Iterate( VFS_TEXT( "testy" ), Shower, VFS_TRUE, 0 ) )
		ReportError();
	wprintf( L"-----\nNum Dirs: %u\nNum Files: %u\nNum Bytes: %u\n",
			 g_uNumDirs, g_uNumFiles, g_uNumBytes );

	if( !VFS_Archive_Extract( VFS_TEXT( "testy" ), VFS_TEXT( "target" ) ) )
		ReportError();

	getch();

	// Shutdown the VFS.
	bSuccess = VFS_Shutdown();
	assert( bSuccess );

	return 0;
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================
